/*----------------------------------------
 * iservcat.c
 *
 * Data which is piped to this process is
 * sent out a tcp connection.
 *----------------------------------------
 */

/*
 * $RCSfile: iservcat.c,v $
 * $Revision: 1.3 $
 * $Date: 1998/11/20 18:57:35 $
 *
 * $Header: //Iservd/iservcat.c#3 $
 * 
 * $Log: iservcat.c,v $
 * Revision 1.3  1998/11/20 18:57:35  chuckh
 * Version number updated
 *
 * Revision 1.2  1998/08/05 19:57:59  chuckh
 * Got rid of some stuff that was placed here accidentally.
 * It wasn't needed anyway.
 *
 * Revision 1.1  1998/07/31 21:01:12  chuckh
 * Initial revision
 *
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/wait.h>
#include <varargs.h>
#include <time.h>
#include <sys/param.h>

/*--------------------------------------------------------------------
 * Revision History:
 *
 * 9/15/94	MAG	Re-written to use getopt(), additional command
 *			line options added. 
 *			Changed freads to reads: the latter are more 
 *			intuitive when standard input is a terminal.
 * 1/27/95	MAG	Tolerates writes with reduced return values.
 *			Additional debugging code.
 *
 * 2/06/95	MAG	Added a command-line argument (like iservd has)
 *			to do an unconditional sleep after all data
 *			is sent, before other measuers.
 * Porting notes:
 *
 * This code is much simpler than iservd, so I have not broken the
 * iservcat source into modules as I have done with that source.
 * In particular, iservcat is not concerned with error logging,
 * nor with pseudo-tty devices, nor with bidirectional data transfer,
 * nor even with survival in the face of error conditions. While iservd
 * must recover from all but the most egregious conditions, (as befits
 * its status as a daemon process), an iservcat process is invoked explicitly
 * when needed so its death affects only the immediate family.
 *-----------------------------------------------------------------------
 */

#define VERSION "2.1"

#include "isysinc.h"	/* System-dependent includes */

#ifdef SVR4
#	define bzero(a,b)	memset(a,0,b)
#	define bcopy(a,b,c)	memcpy(b,a,c)
#endif

#define BUFFER_SIZE	4096
#define LINGER_DELAY	30000

#if (SCO || SVR4)		 /* linger didn't work on sco or svr4 */
#define USER_LINGER	3
#else
#define USER_LINGER	0
#endif

/*
 * Definitions for the TELNET protocol.
 */

#define	DATA	1		
#define	IAC		255		/* interpret as command: */
#define DONT	254
#define DO		253
#define WONT	252
#define WILL	251
#define TELOPT_BINARY	0
#define TELOPT_TM		6
#define TEL_BUF			3		/* size of telopt message 3 bytes */


/* Command-line options */

int binary_mode = 1;	/* defaults to binary telnet mode */
int old_version = 0;	/* defaults to use telnet mode */
int retry_time  = 5;	/* Retry defaults to 5 seconds */
int retry_limit = 128;	/* Retry defaults to 128 attempts */
int debug_mode  = 0;	/* Defaults to off */
int cooking     = 0;	/* Defaults to no cooking, otherwise cooks output  */
int end_delay   = 0;	/* Optional sleep before timing mark sent */


int wait_response();
int send_endmark();
int send_telopt();

#ifndef IS_SCO	/* SCO UNIX declares this in unistd.h, differently */
#ifdef U7	/* SCO UNIXWARE 7 declares this differently too*/
extern int getopt ARGS((int, char *const*, char const*));
#endif	/* U7 */
#else	/* !IS_SCO */
extern int getopt ARGS((int, char**, char*));
#endif	/* IS_SCO */

extern char *optarg;
extern int optind, opterr;

static int checkNumber ARGS((int, int *, char *));

int
start_connection (remote_host, remote_port)
struct hostent *remote_host;
unsigned short remote_port;
{
	struct sockaddr_in sin;
	int sock_fd;
	int retry;

	
	bzero ((char *)&sin, sizeof (sin));
	sin.sin_family = AF_INET;
	
	sin.sin_port = htons (remote_port);
	bcopy (remote_host->h_addr, (caddr_t)&sin.sin_addr,
		remote_host->h_length);
	
	retry = 0;
	while (retry++ < retry_limit)
	{
		/* Note that if the connect fails, we really need in the
		 * general case to close and re-open the socket.
		 * And when _not necessary, it still doesn't hurt.
		 */
		sock_fd = socket (AF_INET, SOCK_STREAM, 0);
		if (sock_fd < 0)
		{
			perror ("socket");
			exit (1);
		}
		if (connect (sock_fd, &sin, sizeof sin) >= 0)
		{
			if(debug_mode)fprintf(stderr,"Connected\n");
			return sock_fd;
		}

		close(sock_fd);
		if (errno != ECONNREFUSED && errno != EADDRINUSE)
		{
			break;
		}
		if ((retry % 10)  == 1)
			perror("Retrying Connect");
		sleep (retry_time);
	}

	perror ("connect");
	exit (1);
	/*NOTREACHED*/

}

void
send_file (fp, sock_fd)
FILE *fp;
int sock_fd;
{
	char buffer[BUFFER_SIZE];
	char encoded_buffer[BUFFER_SIZE*2];
	char *outptr;
	int fd;
	int rbytes;
	int to_read;
	int writ;
	int was_writ;
	
	/* If we are cooking output, the expansion ratio might be bigger */
	/* So we will read less */

	to_read = cooking ? ((sizeof (encoded_buffer)) / 8) : sizeof (buffer);
	
	if (old_version && !cooking)
		outptr = buffer;
	else
		outptr = encoded_buffer;

	fd = fileno(fp);
	while ((rbytes = read(fd, buffer, to_read)) > 0)
	{
		if(debug_mode)fprintf(stderr,"%d bytes read\n", rbytes);
			
		if (!old_version)
		{
			if(cooking)
			{
				rbytes = telnet_cook(buffer,encoded_buffer,rbytes);
			}
			else
			{
				rbytes = telnet_encode(buffer,encoded_buffer,rbytes);
			}
			
		}
		else if (cooking)
		{
			rbytes = only_cook(buffer, encoded_buffer, rbytes);
		}

		
		for(was_writ = 0;;)
		{
			/* Attempt to write whatever hasn't been */
			/* Any real error is fatal */

			if((writ = write(sock_fd, 
				&outptr[was_writ], rbytes - was_writ)) < 0)
			{
				perror ("write");
				exit (1);
			}
			/* In case less was written than planned,
			 * don't sweat it.  Just catch up.  Since we are
			 * supposed to do non-blocking writes, we would
			 * not expect a reduced count, but on some systems
			 * a resource deficiency might cause early return.
			 * Another possibility that we _aren't checking,
			 * is that on some systems a resource crunch moght
			 * lead to a return -1 with errno=EDEADLK or 
			 * similar.  We'll believe _that when we see it.
			 */

			was_writ += writ;
			if(debug_mode)fprintf(stderr,"%d bytes written\n", writ);
			if(was_writ >= rbytes) break;

			sleep(1);	/* No sense retrying _real soon */
		}	       
	}
}

static int
checkNumber(min, dest, source)
	int min;
	int *dest;
	char *source;
{
	if (!isdigit(*source) && *source != '-')
	{
		return 0;
	}

	*dest = atoi(source);
	if (*dest < min ) return 0;
	return 1;
}

char *pName = "Undefined";

static void
usage()
{
	fprintf(stderr, "%s: Version %s\n", pName, VERSION);
	fprintf(stderr,
		"usage: %s [options] <host> <tcp_port> [<-> | <file> ...]\n",
		pName);
	fprintf(stderr,"options: -o         Old IntelliServer compatibility\n");
	fprintf(stderr,"         -a         Use ASCII Telnet, not binary\n");
	fprintf(stderr,"         -c         NL->CRNL and expand tabs\n");
	fprintf(stderr,"         -r limit   Limit on connection retries\n");
	fprintf(stderr,"         -d delay   Seconds between connection retries\n");	fprintf(stderr,"         -x         Enable debugging output\n");
	fprintf(stderr,"Host and tcp_port are required\n");
	fprintf(stderr,"If no files are listed, output is piped from standard input\n");
	fprintf(stderr,"Otherwise files ('-' representing standard input) are\n");
	fprintf(stderr,"send in the order listed.\n");
	exit (1);
}



int
main (argc, argv)
int argc;
char **argv;
{
	struct hostent *remote_host;
	struct linger	delay_close;
	unsigned short remote_port;
	int sock_fd;
	FILE *fp;

	int ovalue;

	pName = argv[0];
	while((ovalue = getopt(argc, argv, "acd:or:w:x")) != EOF)
	{
		switch(ovalue)
		{
		case 'd':	/* Seconds 'tween retries: -d NNN */
			if (!checkNumber(1, &retry_time, optarg))
			{
				fprintf(stderr,"%s: bad delay after -d: %s\n",
					pName, optarg);
				usage();
			}
			break;

		
		case 'a':	/* Disable binary mode: -a */
			binary_mode = 0;
			break;

		case 'c':	/* Enable ONLCR & TAB3 output cooking */
			cooking = 1;
			break;
		case 'o':	/* Old iserve compat mode: -o */
			old_version = 1;
			break;

		case 'r':	/* Retry limit: -r NNN */
			if (!checkNumber(-65000, &retry_limit, optarg))
			{
				fprintf(stderr,"%s: bad limit after -r: %s\n",
					pName, optarg);
				usage();
			}

		case 'w':	/* End delay: -w NNN */
			if (!checkNumber(0, &end_delay, optarg))
			{
				fprintf(stderr,"%s: bad delay after -w: %s\n",
					pName, optarg);
				usage();
			}
		case 'x':	/* Enable debugging output: -x */
			debug_mode = 1;
			break;

		default:
			fprintf(stderr,"%s: bad option: -%c\n",
				pName, ovalue);
			usage();
		}
	}

	/* If we are not a telnet connection (and therefore cannot send
	 * a timing mark), this defaults to 3 seconds.
	 */

	if (old_version)
	{
		binary_mode = 0;
		if(end_delay == 0) end_delay = USER_LINGER;
	}
	if(debug_mode)
	{
		fprintf(stderr, 
			"Debugging on, %s\n%d attempts at %d seconds, ",
			cooking ? "NL->CRNL & Expand tabs, " : "",
			retry_limit, retry_time);
		if(old_version)
		{
			fprintf(stderr, "Non-Telnet mode\n");
		} else if (binary_mode) 
		{
			fprintf(stderr,"Binary Telnet mode\n");
		} else
		{
			fprintf(stderr,"ASCII Telnet mode\n");
		}
	}

	/* After these arguments have been disposed of, there must
	 * remain at least a host and a tcp_port. Go pick them up
	 * or else complain about it.  Then the remaining arguments
	 * (if any) are file names.
	 */

	if (optind < argc)
	{
		/* get IP address for hostname */
		if ((remote_host = gethostbyname (argv[optind])) == NULL)
		{
			fprintf (stderr, "Unknown host %s\n", argv[optind]);
			exit (1);
		}
		if(debug_mode)fprintf(stderr,"IntelliServer: %s\n", argv[optind]);
		optind++;
	}
	else
	{
		usage();
	}

	if (optind < argc)
	{
		remote_port = (unsigned short) atoi (argv[optind]);
		if (remote_port <= 0)
		{
			fprintf (stderr, "Port number must be greater than 0\n");
			usage();
		}
		if(debug_mode)fprintf(stderr," TCP port: %s\n", argv[optind]);
		optind++;
	}
	else
	{
		usage();
	}
	
	sock_fd = start_connection (remote_host, remote_port);

	if (binary_mode) /* send intelliserver binary_mode message */
		send_telopt(sock_fd,DO,TELOPT_BINARY);

	if (optind >= argc)
	{
		if(debug_mode)fprintf(stderr,"Standard Input\n");
		send_file (stdin, sock_fd);
	}
	else
	{
		while (optind < argc)
		{
			if (*argv[optind] == '-')
			{
				if(debug_mode)fprintf(stderr,"Standard Input\n");
				send_file (stdin, sock_fd);
			}
			else
			{
				if ((fp = fopen(argv[optind], "r")) == NULL)
				{
					perror (argv[optind]);
				}
				else
				{
					if(debug_mode)fprintf(stderr,"File: %s\n",
						argv[optind]);
					send_file (fp, sock_fd);
					fclose (fp);
				}
			}
			optind++;
		}
	}

	/* Impose end_delay if defined */

	if(end_delay)
	{
		sleep(end_delay);
	}

	/* set linger on close */
	delay_close.l_onoff = 1;		/* turn on linger*/
	delay_close.l_linger = LINGER_DELAY;	/* delay in seconds*/
	setsockopt(sock_fd, SOL_SOCKET, SO_LINGER, 
			(char *)&delay_close, sizeof(delay_close));

	if (!old_version)
	{
		send_endmark(sock_fd);
	}

	close (sock_fd);
	exit (0);
}

/* 
 * Change IAC's in data to IAC,IAC sequence
 */
int
telnet_encode(source, dest, len)
	unsigned char *source;
	unsigned char *dest;
	int len;
{
	unsigned char *ndx;

	ndx = dest;

	for (; len; len--)
	{
		if ((*source) == IAC)
		{
			*dest++ = IAC;
		}
		*dest++ = *source;

		/* For compatibility with SUNS */
		if ((*source == '\r') && !binary_mode)
		{
			if ((len > 1) && (*(source+1) != '\n'))
				*dest++ = 0;	/* Insert a NULL after cr*/
		}

		source++;
	}
	return(dest - ndx);
}

int
telnet_cook(source, dest, len)
	unsigned char *source;
	unsigned char *dest;
	int len;
{
	unsigned char *ndx;
	int col;

	ndx = dest;
	col = 0;

	for (; len; len--)
	{
		switch(*source)
		{
		case IAC:
			*dest++ = IAC;
			*dest++ = *source++;
			break;

		case '\b':
			if(col > 0) col--;
			*dest++ = *source++;
			break;

		case '\r':
			col = 0;
			/* For compatibility with SUNS */
			if(!binary_mode && len > 1 && (*(source+1) != '\n'))
			{
				*dest++ = 0;
			}
			*dest++ = *source++;
			break;

		case '\n':
			col = 0;
			*dest++ = '\r';
			*dest++ = *source++;
			break;

		case '\t':
			*dest++ = ' ';
			while((++col) & 7)
			{
				*dest++ = ' ';
			}
			source++;
			break;
		default:
			if(*source >= 0x20 && *source < 0x7f) col++;
			*dest++ = *source++;
			break;
		}
	}
	return(dest - ndx);
}

int
only_cook(source, dest, len)
	unsigned char *source;
	unsigned char *dest;
	int len;
{
	unsigned char *ndx;
	int col;

	ndx = dest;
	col = 0;

	for (; len; len--)
	{
		switch(*source)
		{
		case '\b':
			if(col > 0) col--;
			*dest++ = *source++;
			break;

		case '\n':
			col = 0;
			*dest++ = '\r';
			*dest++ = *source++;
			break;

		case '\t':
			*dest++ = ' ';
			while((++col) & 7)
			{
				*dest++ = ' ';
			}
			source++;
			break;
		default:
			if(*source >= 0x20 && *source < 0x7f) col++;
			*dest++ = *source++;
			break;
		}
	}
	return(dest - ndx);
}

int
send_telopt(fd,command,telopt_code)
	int fd;
	unsigned char command;
	unsigned char telopt_code;
{
	unsigned char message[TEL_BUF];

	message[0] = IAC;
	message[1] = command;
	message[2] = telopt_code;

	return(write(fd,message,3));
}

/*
 * SCO and some SVR4 will not linger long enough to put all data on the
 * net. Sending this mark and waiting for a response guarantees that the
 * file has been sent
 */
int
send_endmark(fd)
	int fd;
{
	if(send_telopt(fd,DO,TELOPT_TM) > 0)
	{
		if(debug_mode)fprintf(stderr, "TM?\n");	
		return(wait_response(fd,TELOPT_TM));  
	}
}

int
wait_response( fd, telopt_code)
	int fd;
	unsigned char telopt_code;
{
	int len;
	int cdata;
	int state = DATA;
	unsigned char response[TEL_BUF];
	unsigned char *ndx;

				/* really expect 3 byte packets */
	while((len = read(fd,response,TEL_BUF)) >= 0)
	{
		ndx = response;
		while(len > 0)
		{
			cdata = *ndx++;
			len--; 
			switch(state)
			{
				case DATA:
					switch(cdata)
					{
						case IAC:
							state = IAC;
							break;
						/*default - don't process data*/
					}
					break;
				case IAC:
					switch(cdata)
					{
						case IAC:
							/*don't process data*/
							state = DATA;
							break;
						case DO:
						case DONT:
						case WILL:
						case WONT:
							state  = cdata;
							break;
						/*default - protocol broken */
					}
					break;
				case DO:
				case DONT:
					state = DATA;
					break;
				case WILL:
				case WONT:
					state = DATA;
							/* received response */
					if (cdata == telopt_code)
					{
						if(debug_mode)fprintf(stderr, "TM!\n");	
						return(1);
					}
					break;
			}
		}
	}
	return(0);
}

