/*---------------------------------------------------------------
 * iservcon.c
 *
 * iservd functions that have to do with establishing
 * a socket connection to the server.
 *
 * Compile options:
 *	-DSIGCHLD_EXIT	parent will exit if child dies.
 *			useful if init is starting iservd.
 *---------------------------------------------------------------
 */
/*
 * $RCSfile: iservcon.c,v $
 * $Revision: 1.2 $
 * $Date: 1998/09/23 15:43:28 $
 *
 * $Header: //Iservd/iservcon.c#2 $
 * 
 * $Log: iservcon.c,v $
 * Revision 1.2  1998/09/23 15:43:28  chuckh
 * Changed cpp directive for remote shell command for sco
 *
 * Revision 1.1  1998/07/31 21:02:57  chuckh
 * Initial revision
 *
 */


/*---------------------------------------------------------------
 * Revision History:
 *
 *  5 August  94	MAG	First Draft
 * 30 January 95	MAG	SUNOS requires sigaction if
 *				signals are to interrupt read slow device.
 *
 *---------------------------------------------------------------
 */

#include "isysinc.h"	/* System-dependent includes */

#include "iservopt.h"
#include "iservlog.h"
#include "iservtel.h"
#include "iservcon.h"

/*
 * set the remote shell and its path depending on the UNIX
 * currently its been defined for HP, NCR, MIPS and LINUX. For
 * others rsh is being used. 
 * DO NOT FORGET TO VERIFY THIS FOR YOUR HOST MACHINE!!!!!!!
 */

#if defined(HP) || defined(NCR)
#define	REMSH	"/usr/bin/remsh"
#else	/* not HP or NCR */

#if defined(MIPS)
#define	REMSH	"/usr/net/remsh"
#else	/* not HP, NCR or MIPS */

#if defined(LINUX) || defined(U7)
#define	REMSH	"/usr/bin/rsh"
#else	/* not HP, NCR or MIPS, LINUX */

#if defined(IS_SCO)
#define	REMSH	"/usr/bin/rcmd"
#else	/* others */

/* this will create a warning which hopefully make people double check it */

#define	REMSH	"/usr/bin/remsh"		/* AIX */
#define	REMSH	"/bin/rsh"

#endif /* IS_SCO */

#endif /* LINUX */

#endif	/* MIPS */

#endif	/* HP || NCR */

static int	con_try = 0;


/*---------------------------------------------------------------
 * intrSignal(sig, func)
 *
 * For OS who do not have SA_INTERRUPT defined, this just
 * does a signal, because _these POSIX-compliant ones will
 * terminate read of slow device on receipt of signal.
 * The older BSD systems which restart these reads allow
 * you to use sigaction (but not signal) to force the
 * reads to be interrupted.
 *
 * The two signals we want to ensure will interrupt
 * a "slow-device" read are SIGALRM and SIGPIPE.  The
 * parent process that reads the host and writes the
 * server would be reading from the pseudo-tty master
 * on an idle connection.  So SIGALRM would need to
 * interrupt the read for an idle connection to time 
 * out on a non-permanent connection.  And since the
 * child would send a SIGPIPE to parent when _it sees
 * that the connection is down, we need for SIGPIPE to
 * interrupt the read as well.  (The child would see
 * the SIGPIPE anyway, because _it is reading the network
 * socket, _not a "slow device".
 *---------------------------------------------------------------
 */

#if defined(SA_INTERRUPT)

void
intrSignal(sig, func)
int sig;
void (*func)();
{
	struct sigaction acter;

	memset(&acter.sa_mask, 0, sizeof(acter.sa_mask));
	acter.sa_handler = func;
	acter.sa_flags = SA_INTERRUPT;
	sigaction(sig, &acter, 0);

}

#else	/* SA_INTERRUPT not defined */

void
intrSignal(sig, func)
int sig;
void (*func)();
{
	signal(sig, func);
}

#endif	/* SA_INTERRUPT defined or not */



/*---------------------------------------------------------------
 * gotpipe()
 *	Returns true iff there has been a SIGPIPE
 *
 * servpipe()
 *	The pipe (and alarm) handler. 
 *---------------------------------------------------------------
 */

static int GotPipe;

int
gotpipe()
{
int t = GotPipe;

	GotPipe = 0;
	return t;
}

static void
servpipe()
{
	wakeTelnet();
	GotPipe = 1;
	intrSignal(SIGPIPE, servpipe);
}



/*---------------------------------------------------------------
 * deathOfChild()
 * restarts dead child if necessary
 * often used when init is starting/restarting iservds
 *---------------------------------------------------------------
 */

void
deathOfChild()
{
	extern int ReaderPid;
	int status;
	int result;

	do_log(IL_NOTICE,"Child process died, ReaderPid=%d", ReaderPid);

	result = wait(&status);		/* prevents zombies */

	do_log(IL_NOTICE,"wait()=%d errno=%d status=%02xh,%02xh",
		result, errno, status >> 8, status & 0xFF);

	/* Rearm the SIGCHLD signal */
	signal(SIGCHLD, deathOfChild);

	/* ReaderPid is set when the parent iservd has connected
	 * to the IS and forked a child to read from the network.
	 * On NCR machines we get a death of child from grantpt.
	 */
	if (ReaderPid == 0)
		return;

	ReaderPid = 0;

#ifdef SIGCHLD_EXIT

	/* in this case we want to let init restart everything */

	do_log(IL_NOTICE,"Exiting");

	exit(-1);
#else

	/* restart the child that died */

	forkReader();
	
	return;
#endif
}



/*---------------------------------------------------------------
 * resolveName(pcon)
 * 
 * Attempts to resolve the host name.
 * Returns 1 if successful, 0 otherwise.
 * Does not automatically log if unsuccessful,
 * the name server may be temporarily down or
 * some such.
 *---------------------------------------------------------------
 */

static int
resolveName(pcon)
coninfo_t *pcon;
{
struct hostent *phost;

	if (!(pcon->conFlags & CON_RESOLVED))
	{
		if ((phost = gethostbyname(pcon->conName)) == 
			(struct hostent *) 0)
		{
			return 0;
		}
	
		/* Copy entire host structure, and recopy the address */

		memcpy(&pcon->conHost, phost, sizeof (struct hostent));

		memcpy(&pcon->conIn.sin_addr, pcon->conHost.h_addr, 4);

		pcon->conFlags |= CON_RESOLVED;
	}
	return 1;
}



/*-------------------------------------------------------------------
 * initConnect(pcon, opt)
 * 
 * Initializes the connection structure at pcon
 * based on information in the option structure at
 * opt.
 *
 * Designed for single use (i.e., assuming the socket ain't already
 * open.)
 *
 * Return 1 if ok.
 * Returns 0 of not ok (and will have logged the error if necessary)
 *----------------------------------------------------------------------
 */

int
initConnect(pcon, opt)
coninfo_t *pcon;
servopts_t *opt;
{

	memset((char *) pcon, 0, sizeof (coninfo_t));

	pcon->conPort		= opt->svoPort;
	pcon->conTotal		= opt->svoRetryLimit;
	pcon->conDelay		= opt->svoRetryDelay;
	pcon->conIn.sin_family	= AF_INET;
	pcon->conIn.sin_port	= htons (opt->svoPort);

	if (pcon->conTotal < 0) pcon->conTotal = 0;
	if (pcon->conTotal > 30000) pcon->conTotal = 30000;

	if (opt->svoFlag & SVO_BINARY) pcon->conFlags |= CON_BINARY;
	if (opt->svoFlag & SVO_PERM) pcon->conFlags |= CON_PERM;
	if (opt->svoFlag & SVO_HANGUP) pcon->conFlags |= CON_HANG;
	if ((opt->svoFlag & SVO_OLD)) 
	{
		/* Sleep before close only enforced for non-telnet */
		pcon->conCloseZZ = opt->svoCloseDelay;
		pcon->conFlags &= ~CON_BINARY;
	}
	else
	{
		pcon->conFlags |= CON_TELNET;
	}

	memcpy(pcon->conName, opt->svoServer, sizeof (pcon->conName));

	resolveName(pcon);	/* Attempt to resolve the name now */
				/* If fails, no sweat until we need a conn. */

	pcon->conFlags |= CON_READY;
	return 1;
}



/*---------------------------------------------------------------
 * endConnection(pcon)
 * Ends a connection.
 * If connection is not already marked as lost, we will
 * set appropriate options to prevent the loss of data.
 * If connection is marked as lost, we will just close the
 * socket and get things ready for re-connection later.
 * (If no connection seems to be active, we just return)
 *---------------------------------------------------------------
 */

#define MAX_LINGER 30000

void
endConnection(pcon)
coninfo_t *pcon;
{
struct linger ling;

	if(!(pcon->conFlags & CON_OPEN)) return;

	if ((pcon->conFlags & CON_LOST))
	{
		do_log(IL_NOTICE,"Connection Dropped by %s", pcon->conName);
	}
	else
	{
		do_log(IL_NOTICE,"Ending connection to %s", pcon->conName);

		if(pcon->conFlags & CON_TELNET)
		{
			/* If a telnet connection, uset the telnet marker */

			waitTelnet(pcon->conFd);
		}
		else
		{
			/* If a non-telnet connection, set LINGER.
			 * I did not set it earlier (I suppose) because
			 * I don't want _unexpected closes to block (i.e.,
			 * because of exits, killing the job, etc...)
			 */

			do_log(IL_DEBUG,"Setting SO_LINGER");

			ling.l_onoff = 1;
			ling.l_linger = MAX_LINGER;

			setsockopt(pcon->conFd, SOL_SOCKET, SO_LINGER,
				(char *)&ling, sizeof(ling));
		}
	}
	
	/* And, in the unlikely event someone used the -w option... */

	if(pcon->conCloseZZ)
	{
		do_log(IL_DEBUG, "Waiting %d seconds", pcon->conCloseZZ);
		sleep(pcon->conCloseZZ);
	}

	close(pcon->conFd);

	pcon->conFlags &= ~(CON_OPEN | CON_LOST);

	con_try = 0;
}



/*---------------------------------------------------------------
 * markLost(pcon, message)
 * Marks the connection as lost, and sends
 * a debugging message to the syslog or the 
 * logfile (provided debugging is enabled)
 * If the message is a null pointer, then no
 * message is sent.
 *---------------------------------------------------------------
 */

void
markLost(pcon, message)
coninfo_t *pcon;
char *message;
{
	if(pcon->conFlags & CON_LOST) return;

	if (message)
	{
		do_log(IL_DEBUG, message);
	}
	pcon->conFlags |= CON_LOST;
}



/*-----------------------------------------------------------------
 * startConnection(pcon)
 * Attempts to start a connection 
 * Assumes that the connection has not been made yet: if there seems
 * already to be one, we just return.
 * called by:
 * hostToServer() -> connectAndFork() -> attemptConnect() -> startConnection()
 * Returns CON_OK if connection has been established,
 * Returns CON_BUSY if connection failed for non-fatal reasons.
 * Returns CON_ERROR if connection failed for fatal reasons.
 *----------------------------------------------------------------
 */

#define CONN_OK		1
#define CONN_BUSY	2
#define CONN_ERROR	3

static int
startConnection(pcon)
coninfo_t *pcon;
{
int on = 1;
char	cmd[100];
char	port;
int	rval;

	if (pcon->conFlags & CON_OPEN) return CONN_OK;	/* Already open */

	if (con_try == 200)
		con_try = 4;
	else
		con_try++;

	if (!resolveName(pcon))
	{
		do_log(IL_DEBUG, "Could not resolve host name %s",
			pcon->conName);
		return CONN_BUSY;	/* Yet another reason to retry later */
	}

	if((pcon->conFd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		do_log(IL_ERROR, "Cannot open a socket %d", errno);
		return CONN_ERROR;
	}

	/* Connect will bind automatically, so an explicit bind is 
	 * not required here.
	 */

	if(connect(pcon->conFd, (struct sockaddr *)&pcon->conIn, 
		sizeof (struct sockaddr_in)) < 0)
	{
		close(pcon->conFd);

		/* Most errors simply mean we should retry.
		 * But some are serious, and no amount of retrying will
		 * fix the problem.
		 */
		switch(errno)
		{
		case ECONNREFUSED:	

			/* This could be due to network going down and
			 * iservd unable to close the socket connection.
			 * As a result, when the iservd comes up the next
			 * time, the connection will be refused from the
			 * intelliserver as the old connection is still
			 * active on the intelliserver. 
			 * By doing a remote kill port on the intelliserver
			 * all the exsisting connections are closed, and
			 * the iservd will be able to connect on the next
			 * attempt for connection.
			 */
			
			do_log(IL_DEBUG, "Connection refused");

			/* For the first three attempts we wont do remote
			 * kill port even if -I option is specified, because
			 * some networks could be slow.
			 */
			if (con_try < 3)
				return CONN_BUSY;

			if (rmt_cmd_flag & REMOTE_KILL_PORT)
			{
				port = ntohs(pcon->conIn.sin_port) - 9000;
				sprintf(&cmd[0], "%s %s %s %d", (char *)REMSH,
					 pcon->conName, "kill port ", port);

				do_log(IL_DEBUG, "remote command : %s \n", cmd);		

				rval = system(cmd);
				if ((rval == 127) || (rval == -1))
				{
					do_log(IL_DEBUG, "system() failed for cmd %s\n",
						cmd);
				}
			}
			return CONN_BUSY;

		case ETIMEDOUT:
			do_log(IL_DEBUG, "Connection timed out");
			return CONN_BUSY;
		case EADDRINUSE:
		case EINTR:
			do_log(IL_DEBUG, "Cannot connect: errno %d", errno);
			return CONN_BUSY;	/* We should try again */

		case ENETUNREACH:
			do_log(IL_ERROR, "Network unreachable");
			return CONN_ERROR;
		default:		/* Anything else is freak */
					/* Send error message */
			do_log(IL_ERROR, "Cannot connect: unexpected error %d", errno);
			return CONN_ERROR;
		}
	}	

	do_log(IL_NOTICE,"Connected to %s", pcon->conName);
		
	gotpipe();	/* Clear pipe-signal flag */

	intrSignal(SIGPIPE, servpipe);

	if (setsockopt(pcon->conFd, SOL_SOCKET, 
		SO_KEEPALIVE, &on, sizeof(on)) < 0)
	{
		do_log(IL_ERROR, "could not set KEEPALIVE option");
	}

	if(pcon->conFlags & CON_BINARY)
	{
		if(setBinary(pcon->conFd) <= 0)
		{
			do_log(IL_ERROR, "could not set binary telnet option");
		}
	}
	pcon->conFlags |= CON_OPEN;

	return CONN_OK;
}	



/*---------------------------------------------------------------
 * attemptConnect(pcon)
 *
 * Attempts a connection with the host,
 * retrying for a specified number of times.
 *
 * Return values: 1 = successful
 * 0 = either the retries were exhausted, or
 *	a "fatal" error was encountered.
 *
 * Note that the try count is reset at 31000, which is greater
 * than the maximum number of retries allowed.
 *---------------------------------------------------------------
 */

static unsigned int con_delay;

int
attemptConnect(pcon)
coninfo_t *pcon;
{
	for(pcon->conTry = 1;
		pcon->conTotal == 0 ||  (pcon->conTry <= pcon->conTotal);
		pcon->conTry++)
	{
		switch(startConnection(pcon))
		{
		case CONN_OK:
			return 1;

		case CONN_ERROR:
			sleep(10);	/* must be kernel problem! */
			return 0;

		case CONN_BUSY:
		default:
			break;
		}	
		
		/* for the first three attempts, we are going to ignore
		 * the retry times specified by the user, instead we 
		 * sleep for 1 second before attempting to connect again.
		 */
		if (con_try < 3) {
			do_log(IL_ERROR, "sleeping  1 second before retrying");/* TODO - remove */
			sleep(1);
			continue;
		}
		/*
		 * a customer want to make sure log files dont grow without
		 * limit so this tapers it off slowly over time, change it
		 * if you want
		 */

		con_delay += 2;

		do_log(IL_ERROR, "sleeping  %d seconds before retrying", con_delay);/* TODO - remove */
		sleep(con_delay);


		if(pcon->conDelay) {
			do_log(IL_DEBUG, "sleeping  %d seconds before retrying", pcon->conDelay);/* TODO - remove */
			sleep(pcon->conDelay);
		}
	}

	do_log(IL_ERROR, "%d attempts: connection failed",
		pcon->conTotal);

	/* If this is a permanent connection and we are here, it
	 * means he has explicitly specified the -r option to limit
	 * the retries.  If we return immediately, we will just manage
	 * to start trying to connect again, so we insert the completely
	 * gratuitious (but documented) 60-second delay before we
	 * try again.
	 */

	if(pcon->conFlags & CON_PERM) sleep(60);

	return 0;
}




