/*---------------------------------------------------------------
 * iservlog.c
 * 
 * Procedures for handling all logging
 * (error messages, debugging output, etc)
 * for iservd
 *
 * Compile switches:
 * -DSYSLOG	for those systems which do support
 * 		the syslog call.
 * -DMIPS	fiddle include files for MIPS RISCOS
 *		(e.g., MIPS Magnum 3000)
 *---------------------------------------------------------------
 */
/*
 * $RCSfile: iservlog.c,v $
 * $Revision: 1.2 $
 * $Date: 1998/08/04 20:18:57 $
 *
 * $Header: //Iservd/iservlog.c#1 $
 * 
 * $Log: iservlog.c,v $
 * Revision 1.2  1998/08/04 20:18:57  chuckh
 * The process is now properly identified as iservdc
 *
 * Revision 1.1  1998/07/31 21:11:22  chuckh
 * Initial revision
 *
 */


/*---------------------------------------------------------------
 * Revision History:
 * 
 * 07/28/94	MAG	First draft
 *---------------------------------------------------------------
 */

#ifdef SYSLOG

#if defined(MIPS)
#include <bsd/syslog.h>
#elif defined(IS_SCO)
#include <sys/syslog.h>
#elif 1	/* Not Mips or SCO (Mips doesn't like "else" here!) */
#include <syslog.h>
#endif	/* Mips, SCO, or other */

#endif	/* SYSLOG */

#include "isysinc.h"
#include "iservopt.h"
#include "iservlog.h"


static int fileOpts;	/* What kinds of messages are logged locally? */
static int sysOpts;	/* What kings of messages are syslogged? */
static char formatOpt;

static char delimit;



/*---------------------------------------------------------------
 * init_log(opts)
 * 
 * Initializes all the logging options based
 * on the user-selected options.
 *
 * Returns 0: unable to open the log file
 * Returns 1: ok
 *---------------------------------------------------------------
 */

int
init_log(opts)
servopts_t * opts;
{
int flag = opts->svoFlag;

	fileOpts = 0;
	sysOpts = 0;
	delimit = '-';

	/* Sort out the local log-file issues first */

	/* If they have specified a log file or debugging, then
	 * it is mandatory to have that file. If the file can't be
	 * opened, we return with an error.  If no local log is required,
	 * we will leave stderr as is.
	 */

	if (flag & SVO_LOG)
	{		
		close(2);	/* stderr will be the stream to use */
		if (open(opts->svoLogFile, O_WRONLY|O_CREAT|O_APPEND, 0666) < 0)
		{
			return 0;
		}
	
		fileOpts |= IL_NOTICE | IL_ERROR;
		if (flag & SVO_DEBUG) 
		{
			fileOpts |= IL_DEBUG;
			if (flag & SVO_VIEW)
			{
				fileOpts |= IL_VIEW;
				formatOpt = opts->svoView;
			}
		}
	}

#ifdef SYSLOG

	if (flag & SVO_SYSLOG)
	{
		sysOpts |= IL_NOTICE | IL_ERROR;

		openlog("iservdc", LOG_PID, LOG_DAEMON);

		if (flag & SVO_DEBUG)
		{
			sysOpts |= IL_DEBUG;
			if (flag & SVO_VIEW) sysOpts |= IL_VIEW;
		}

		syslog(LOG_INFO,"sysloging enabled");
	}

#endif /* SYSLOG */


	return 1;
}


/*---------------------------------------------------------------
 * markChild()
 * Called after fork so that subsequent
 * output from _that process will be marked
 *---------------------------------------------------------------
 */

void
markChild()
{
	delimit = '=';	/* Will appear on debugging log thus */
}



/*---------------------------------------------------------------
 * do_log(flavout, fmt, ...)
 *
 * Logs the message to the the syslog if selected, 
 * and/or the local log file is selected.
 * Linefeeds are automatically appended, so should
 * NOT be sent as part of the format string.
 *---------------------------------------------------------------
 */
 
void
do_log(flavour, fmt, va_alist)
int flavour;
char *fmt;
va_dcl
{
va_list	ap;
char tbuffer[4096];	/* More than we need, but best to be safe */
int length;
int priority;
time_t timer;

	if (flavour & fileOpts)
	{
		/* Pick up the time and strip the last char,
		 * presumed to be a NL.
		 */
		time(&timer);
		strcpy(tbuffer, ctime(&timer));
		length=strlen(tbuffer);
		if (length > 1 && tbuffer[length-1] == '\n')
		{
			tbuffer[length-1] = delimit;
		}
		fprintf(stderr, tbuffer);
			
		/* Print the remainder of the message */
		va_start(ap);
		vfprintf(stderr, fmt, ap);
		va_end(ap);
		fprintf(stderr, "\n");
		fflush(stderr);
	}
#ifdef SYSLOG
	if (flavour & sysOpts)
	{
		if(flavour & IL_NOTICE) priority = LOG_NOTICE;
		else if(flavour & IL_ERROR) priority = LOG_ERR;

		priority = LOG_DEBUG;

		/* Note we need to be modest in what we would send,
		 * since vsprintf offers no intrinsic overflow
		 * protection.
		 */

		vsprintf(tbuffer, fmt, ap);
		syslog(priority, tbuffer);
	}
#endif
}	



/*---------------------------------------------------------------
 * traceData
 * write the data to the log destination
 *---------------------------------------------------------------
 */

void
traceData(message, buffer, count)
char *message;
unsigned char *buffer;
int count;
{
char outbuff[80];
int thisTime;
int thatTime;
char *pout;
unsigned char uc;

	if (!((sysOpts | fileOpts) & IL_VIEW)) return;

	/* We must RESIST THE TEMPTATION to depend on sprintf returning
	 * the number of bytes written, i.e. pout += sprintf(...)
	 * because on some BSD-based systems (including SUN), sprintf
	 * is a char pointer and returns the first argument.
	 * So we make the most portable assumption and calculate
	 * the new starting place after each sprintf.  We start the
	 * strlen() where the previous one left off to avoid
	 * geometric growth.
	 */

	while(count)
	{
		pout = outbuff;
		sprintf(pout, message);
		pout += strlen(pout);	/* SEE BIG NOTE ABOVE... */

		thatTime = thisTime = (count < 16) ? count : 16;

		switch(formatOpt)
		{
		case 'd':
			while(thisTime--)	
			{
				sprintf(pout, "%3u ", *buffer++);
				pout += strlen(pout);
			}
			break;
		case 'o':
			while(thisTime--)	
			{
				sprintf(pout, "%03o ", *buffer++);
				pout += strlen(pout);
			}
			break;
		case 'a':
			while(thisTime--)
			{
				uc = *buffer++;
				if (!isalnum(uc))
				{
					sprintf(pout, "%02x ", uc);
				}
				else	
				{
					sprintf(pout, "%c ", uc);
				}
				pout += strlen(pout);
			}
			break;
		default:
		case 'x':
			while(thisTime--)	
			{
				sprintf(pout, "%02x ", *buffer++);
				pout += strlen(pout);
			}
			break;
		}
		do_log(IL_VIEW, outbuff);
		count -= thatTime;
	}
}


