/*---------------------------------------------------------------
 * iservopt.c
 * 
 * iservd routines that are related to the
 * parsing and storage of command-line 
 * arguments and options.
 *
 * Compile Flags:
 *	-DFOREVER	If -t option is not give,
 *			default to forever, else
 *			default to 60 seconds.
 *			Only use this option for streams
 *			OS's that will return a 0 on the
 *			master pseudo read when the slave is
 *			closed.  For OS's that don't do that,
 *			the timeout is the only way to suppose
 *			that the slave port might be closed.
 *---------------------------------------------------------------
 */
/*
 * $RCSfile: iservopt.c,v $
 * $Revision: 1.2 $
 * $Date: 1998/08/04 20:25:33 $
 *
 * $Header: //Iservd/iservopt.c#3 $
 * 
 * $Log: iservopt.c,v $
 * Revision 1.2  1998/08/04 20:25:33  chuckh
 * All boss functionality has been removed
 *
 * Revision 1.1  1998/07/31 21:12:28  chuckh
 * Initial revision
 *
 */


/*---------------------------------------------------------------
 * Revision History:
 * 
 * 07/28/94	MAG	First draft
 * 03/05/95	MAG	-i option added
 *---------------------------------------------------------------
 */

#include <string.h>

#include "isysinc.h"	/* System-dependent includes */
#include "iservopt.h"

/* Used by getopt() */

#ifndef IS_SCO /* Defined by SCO in unistd.h, differently */
#ifdef U7 /* Defined by SCO UNIXWARE differently too */
extern int getopt ARGS((int, char *const*, const char *));
#endif	/* U7 */
#else /* !IS_SCO */
extern int getopt ARGS((int, char**, char*));
#endif /* IS_SCO */

extern char *optarg;
extern int optind, opterr;
 


/*---------------------------------------------------------------
 * debugOpts(opts)
 * displays the contents of the option structure.
 *---------------------------------------------------------------
 */

#ifdef DEBUG
void
debugOpts(opts)
servopts_t *opts;
{
	fprintf(stderr,"Flag = %x RetryDelay = %d ", opts->svoFlag, opts->svoRetryDelay);
	fprintf(stderr,"Timeout = %d RetryLimit = %d\n", opts->svoTimeout, opts->svoRetryLimit);
	fprintf(stderr,"Port = %d CloseDelay = %d\n", opts->svoPort, opts->svoCloseDelay);
	
	fprintf(stderr,"LogFile=%s\n", opts->svoLogFile);
	fprintf(stderr,"Link=%s\n", opts->svoLink);
	fprintf(stderr,"Server=%s\n", opts->svoServer);
	fprintf(stderr,"Error=%s\n", opts->svoError);
	fprintf(stderr,"view=%c\n", opts->svoView);
}
#endif



/*---------------------------------------------------------------
 * defaultOpts(opts)
 * Sets the servopts_t elements at opts to default
 * values.
 *---------------------------------------------------------------
 */

void
defaultOpts(opts)
servopts_t *opts;
{
	memset((char *)opts, 0, sizeof (servopts_t));
	opts->svoFlag = SVO_BINARY;
	opts->svoRetryDelay = 5;	/* seconds */
	opts->svoRetryLimit = 128;	/* times */
					/* But if a permanent connection, this
					 * will default to forever.
					 */
#ifdef FOREVER
	opts->svoTimeout = 0;		/* forever */
#else
	opts->svoTimeout = 60;		/* seconds */
#endif
	opts->svoCloseDelay = 0;	/* seconds */
}



/*---------------------------------------------------------------
 * defaultBoss(boss)
 * Sets the servboss_t elements at boss to default
 * values.
 *---------------------------------------------------------------
 */
void
defaultBoss(boss)
servboss_t *boss;
{
	memset((char *)boss, 0, sizeof(servboss_t));
	strcpy(boss->bosConf, IS_DEFAULT_CONF);
	strcpy(boss->bosLock, IS_DEFAULT_LOCK);
}

/*---------------------------------------------------------------
 * setOptError(opts, string)
 * Helper routine.
 * Sets SVO_SYNTAX in svoFlag, and
 * copies the error string to svoError
 *---------------------------------------------------------------
 */

static void
setOptError(opts, string)
servopts_t *opts;
char *string;
{
	opts->svoFlag |= SVO_SYNTAX;
	strcpy(opts->svoError, string);
}



/*---------------------------------------------------------------
 * rationalOpts(opts, boss)
 * Examines the set of options at opts, reconciling
 * semantics. If there is an inconsistency, the SVO_SYNTAX
 * bit will be set, and the error message will be in
 * svoError
 *---------------------------------------------------------------
 */

void
rationalOpts(opts, bos)
servopts_t *opts;
servboss_t *bos;
{
int flag = opts->svoFlag;

	if (flag & SVO_SYNTAX) return;

	/* Ensure required arguments are present */

	if ((flag & SVO_REQUIRED) != SVO_REQUIRED)
	{
		setOptError(opts, "Incomplete Arguments");
		return;
	}
	
	/* Ignore -v option if -x not present */

	if (!(flag & SVO_DEBUG)) 
	{
		opts->svoFlag &= ~SVO_VIEW;
	}

	/* If they asked for debugging and there is no log file,
	 * the debug output goes into
	 * /tmp/[linkname].debug
	 * (former versions went to ./[linkname].debug which
	 * seems a bit non-deterministic for my liking.
	 */

	if ((flag & SVO_LOG) && !(flag & SVO_LOGOPT) )
	{
		strcpy(opts->svoLogFile, "/tmp/");
		{
			char	*cp;

			/*
			 * Find the last slash in the string.
			 * If a slash was found, copy
			 * everything after the slash.
			 */
			if (cp = strrchr(opts->svoLink, '/'))
			{
				strcat(opts->svoLogFile, cp+1);
			}

			/*
			 * Otherwise take the entire link name
			 */
			else
			{
				strcat(opts->svoLogFile, opts->svoLink);
			}
		}
		strcat(opts->svoLogFile, ".debug");
	}

	/* If it's a permanent connection, and no -r {times} option
	 * was given, then we should default to trying forever.
	 * But if they give a limit, it's their decision.
	 */

	if ((flag & (SVO_PERM | SVO_RETRIES)) == SVO_PERM)
	{
		opts->svoRetryLimit = -1;
	}

	/* If it's a permanent connection, DO NOT TIME OUT on inactivity */
	/* ...even if they _did specify the -t option */

	if (flag & SVO_PERM)
	{
		opts->svoTimeout = 0;		/* forever */
	}
	else if ((flag & SVO_IGCLOSE) && opts->svoTimeout == 0)
	{
		setOptError(opts, "Temp. connection using -i needs -t also");
		return;
	}

	/* Propagate the lock-file name from the argument */
	if (flag & SVO_LOCK)
	{
		strcpy(bos->bosLock, opts->svoLockFile);
		bos->bosLockFd = open(bos->bosLock, O_RDWR|O_CREAT, 0666);
		if (bos->bosLockFd >= 0)
		{
			bos->bosFlag |= BOS_ISLOCK;
		}
	}
}



/*---------------------------------------------------------------
 * checkNumber(min,destination, source)
 * Helper routine for parseOpts.
 * checks string at source against destination,
 * uses atoi to store an int at destination, and
 * returns true if the number qualifies.
 * returns false otherwise.
 *---------------------------------------------------------------
 */

static int
checkNumber(min, dest, source)
int min;
int *dest;
char *source;
{
	if (!isdigit(*source) && *source != '-')
	{
		return 0;
	}

	*dest = atoi(source);
	if (*dest < min ) return 0;
	return 1;
}



/*---------------------------------------------------------------
 * parseOpts(argc, argv, opts)
 * 
 * Given the argc and argv, uses getopt to parse the
 * command-line options and arguments, filling in the
 * servopts_t structure at opts.
 * these svo opts get copied to other structs later.
 * called from mainChild in iservd.c
 *---------------------------------------------------------------
 */

void parseOpts(argc, argv, opts)
int argc;
char ** argv;
servopts_t *opts;
{
int ovalue;

	while((ovalue = getopt(argc, argv, IS_ARGS)) != EOF)
	{
		switch(ovalue)
		{
		case 'c':	/* Initialize Slave port to uncooked input */
				/* But with cooked output */

			opts->svoFlag |= SVO_OCOOK;
			break;

		case 'd':	/* Seconds 'tween retries: -d NNN */
			if (!checkNumber(1, &opts->svoRetryDelay, optarg))
			{
				setOptError(opts,"bad delay after -d");
				return;
			}
			break;

		case 'h':	/* Hang up on disconnect: -h */
			opts->svoFlag |= SVO_HANGUP;
			break;

		case 'i':	/* -i Ignore slave close (timeout only) */
			opts->svoFlag |= SVO_IGCLOSE;
			break;

				/* Passed from parent, not from config file */
		case 'k':	/* name of lock file: -k AAA */
			opts->svoFlag |= SVO_LOCK;
			strncpy(opts->svoLockFile, optarg, SV_MAX);
			opts->svoLockFile[SV_MAX-1] = 0;
			break;

		case 'l':	/* Log file name: -l AAA */
			opts->svoFlag |= SVO_LOGOPT;
			strncpy(opts->svoLogFile, optarg, SV_MAX);
			opts->svoLogFile[SV_MAX-1] = 0;
			break;
		
		case 'a':	/* Disable binary mode: -a */
		case 'm':	/* Disable binary mode: -m (old-style) */
			opts->svoFlag &= ~SVO_BINARY;
			break;

		case 'o':	/* Old iserve compat mode: -o */
			opts->svoFlag |= SVO_OLD;
			break;

		case 'p':	/* Make permanent connection: -p */
			opts->svoFlag |= SVO_PERM;
			break;

		case 'r':	/* Retry limit: -r NNN */
			if (!checkNumber(-65000, &opts->svoRetryLimit, optarg))
			{
				setOptError(opts,"bad retry count after -r");
				return;
			}
			opts->svoFlag |= SVO_RETRIES;
			break;
			
		case 's':	/* Syslogging requested: -s */
			opts->svoFlag |= SVO_SYSLOG;
			break;
	
		case 't':	/* Inactivity timeout: -t NNN */
			if (rmt_cmd_flag & REMOTE_MSR_CMD) {
				setOptError(opts, 
					"-M and -t are mutually exclusive");
				return;
			}
			if (!checkNumber(0, &opts->svoTimeout, optarg))
			{
				setOptError(opts,"bad timeout after -t");
				return;
			}
			break;

		case 'u':	/* Initialize Slave port to uncooked mode */
			opts->svoFlag |= SVO_SRAW;
			break;

		case 'v':	/* View data: -v [a|d|h] */
			opts->svoFlag |= SVO_VIEW;
			opts->svoView = optarg[0];
			break;

		case 'w':	/* Wait on close: -w NNN */
			if (!checkNumber(1, &opts->svoCloseDelay, optarg))
			{
				setOptError(opts,"bad delay after -w");
				return;
			}
			break;

		case 'x':	/* Enable debugging output: -x */
			opts->svoFlag |= SVO_DEBUG;
			break;

		case 'R' :
			/* Enable the RTS up/down mode */
			rmt_cmd_flag |= REMOTE_RTS_CMD;
			break;

		case 'I' :
			/* Enable insist_flag - to kill the port on IS if
			 * connection is refused
			 */
			rmt_cmd_flag |= REMOTE_KILL_PORT;
			break;

		case 'D' :
			/* delete device linked to pty when we exit
			 */
			opts->svoFlag |= SVO_DELETE;
			break;
		
		case 'M' :
			/* Get the MSR value, to enable us to check for
			 * DCD UP/DOWN for modem ports
			 */
			if (opts->svoTimeout) {
				setOptError(opts, 
					"-M and -t are mutually exclusive");
				return;
			}
			if (!checkNumber(1, &opts->svoMsrTimeout, optarg))
			{
				setOptError(opts,"bad timeout after -M");
				return;
			}
			rmt_cmd_flag |= REMOTE_MSR_CMD;
			break;
		default:
			setOptError(opts,"invalid option");
			return;
		}
	}

	/* Now dispose of the remaining arguments */

	for (ovalue = 0; optind < argc; optind++, ovalue++)
	{
		switch(ovalue)
		{
		case 0: /* the device name */
			opts->svoFlag |= SVO_LINK;
			strncpy(opts->svoLink, argv[optind], SV_MAX);
			opts->svoLink[SV_MAX-1] = 0;
			break;
		case 1: /* the intelliserver's name */
			opts->svoFlag |= SVO_SERVE;
			strncpy(opts->svoServer, argv[optind], SV_MAX);
			opts->svoServer[SV_MAX-1] = 0;
			break;
		case 2: /* the 90xx port number */
			if (isdigit(argv[optind][0])) 
			{
				opts->svoFlag |= SVO_PORT;
				opts->svoPort = atoi(argv[optind]);
			} 
			break;
		default:
			setOptError(opts,"extra arguments");
			return;
		}
	}
}



/*---------------------------------------------------------------
 * parseConfigLine(source,dest)
 * 
 * source points to a line of an iservd.conf file.
 * dest points to a string onto which will be catenated
 * a re-formatted version of same, suitable to be passed
 * as a set of command-line arguments. 
 * The caller must insure that there will be enough room in
 * dest for the resulting string to be added.
 * Return values:
 *	PARSED_IGNORE	the line was a comment.
 *	PARSED_ERROR	the line had bad syntax.
 *	PARSED_GOOD	the line was good.
 *
 * The original copy at source will be modified.
 *---------------------------------------------------------------
 */

int
parseConfigLine(source, dest)
char * source;
char * dest;
{
char *pc;
char *parg;
int params;
char *pstart;

	while (isspace(*source)) source++;	/* Ignore leading spaces */

	/* Discard trivial lines */

	switch(*source)
	{
	case 0:
	case '#':
		return PARSED_IGNORE;
	default:
		break;
	}

	/* Line is non-trivial
	 * First pass, convert any white space to a true space, and
	 * if a '#' is found, terminate the string there.
	 * If there is any '-' present, mark its location, because 
	 * that is the start of the options, and we want them first.
	 */

	parg = 0;
	for(pc = source; *pc; pc++)
	{
		if (isspace(*pc))
		{
			*pc = ' ';
		}
		else if (*pc == '#')
		{
			*pc = 0;
			break;
		}
		else if (!parg && *pc == '-')
		{
			parg = pc;
		}
	}

	/* Regardless of how the line ended originally, now it ends
	 * in a null for sure.
	 */

	*pc = 0;
	
	/* If there were options, copy them first */
	if (parg)
	{
		strcat(dest, parg);
		strcat(dest, " ");
		*parg = 0;
	}
	else
	{
		parg = pc;
	}	

	/* What remains, if anything, are the three required
	 * parameters, which now go last. If the first parameter
	 * ends with a colon, remove the colon.
	 */	

	for(pstart = 0, pc = source, params = 0;pc <= parg;pc++)
	{
		if (pc == parg || isspace(*pc))
		{
			if (pstart)
			{
				*pc = 0;
				if(!params && pc[-1] == ':')
				{
					pc[-1] = 0;
				}
				params++;
				strcat(dest, pstart);
				strcat(dest, " ");
				pstart = 0;
			}
		} 
		else
		{
			if (!pstart) pstart = pc;
		}
	}
	if (params != 3) 
	{
		return PARSED_ERROR;
	}
	return PARSED_GOOD;
}
