/*---------------------------------------------------------------
 * iservpsu.c
 * 
 * Iservd routines to find, allocate, open
 * and otherwise manipulate pseudo-tty master
 * and slave ports.
 *
 * This one gets to be heavily conditional
 * based on the type of OS being used.
 *
 * Compile Flags:
 * -DCLONUS	the OS supports Clone-devices
 * -DPUSH_A	Supports Auto-Push
 * -DPUSH_M	Supports Manual-Push
 * -DPUSH_X	No push used.
 * -DAIX32	Variations among AIX versions
 * -DAIX31	old names, 
 * -DMAKE_NODE	make nodes instead of links
 * 
 *---------------------------------------------------------------
 */
/*
 * $RCSfile: iservpsu.c,v $
 * $Revision: 1.2 $
 * $Date: 1998/08/25 15:03:07 $
 *
 * $Header: //Iservd/iservpsu.c#2 $
 * 
 * $Log: iservpsu.c,v $
 * Revision 1.2  1998/08/25 15:03:07  chuckh
 * When a connection was lost, the pty was rebuilt from scratch.
 * However, an auto-push was being done only the first time through.
 *
 * Revision 1.1  1998/07/31 21:13:13  chuckh
 * Initial revision
 *
 */


/*---------------------------------------------------------------
 * Revision History:
 *
 * 29 July 94	MAG	First draft
 * 30 Oct  95   NSH	added mknod stuff
 *---------------------------------------------------------------
 */

#include "isysinc.h"	/* System-dependent includes */
#include <sys/stat.h>

#ifdef PUSH_A	/* PUSH_A: Autopush supported by this os */
# include	<sys/mkdev.h>
# include	<sys/conf.h>
# include	<sys/sad.h>
# include	<netinet/in.h>
# include	<netdb.h>
#endif		/* PUSH_A */

#include "iservopt.h"
#include "iservlog.h"
#include "iservpsu.h"



/*---------------------------------------------------------------
 * Defines to handle fairly trivial name changes
 * between os's that are fairly close.
 *---------------------------------------------------------------
 */

#if defined(AIX32) || defined(AIX31)
#define LINK		symlink
#else
#define LINK		link
#endif

#if defined(AIX32)

#define CLONE_NAME	"/dev/ptc"
#define PTSNAME		ttyname

#else	/* not AIX32 */

#if defined(HP)

#define CLONE_NAME	"/dev/ptym/clone"
#define PTSNAME		ptsname

#else /* not HP, nor AIX32 */

#define CLONE_NAME	"/dev/ptmx"
#define PTSNAME		ptsname

#endif	/* HP */
#endif	/* AIX32 */



/*---------------------------------------------------------------
 * initPsu(psu, opt)
 * Routine to initialize psu structure
 * to default state, using command-line options.
 * copy cmd line opt flags to psu struct
 *---------------------------------------------------------------
 */

void
initPsu(psu, opt) 
servpsu_t *psu;
servopts_t *opt;
{
	memset((char *)psu, 0, sizeof (servpsu_t));

	/*
	 * If the given device name begins with a leading
	 * slash, interpret it as an absolute path name.
	 * Otherwise, prepend a "/dev/" to it to form the
	 * device name.
	 */
	if (opt->svoLink[0] == '/')
	{
		strcpy(psu->psuLink, opt->svoLink);
	}
	else
	{
		strcpy(psu->psuLink, "/dev/");
		strcat(psu->psuLink, opt->svoLink);
	}
	printf("psuLink set to %s\n", psu->psuLink);

	if(opt->svoFlag & SVO_PERM)   psu->psuFlag |= PSU_PERM;
	if(opt->svoFlag & SVO_SRAW)   psu->psuFlag |= PSU_SRAW;
	if(opt->svoFlag & SVO_OCOOK)  psu->psuFlag |= PSU_OCOOK;
	if(opt->svoFlag & SVO_IGCLOSE)psu->psuFlag |= PSU_IGCLOSE;
	if(opt->svoFlag & SVO_DELETE) psu->psuFlag |= PSU_DELETE;
}



/*---------------------------------------------------------------
 * openMaster(psu)
 * Routine to find and open a master psudo-tty
 * Only valid to use this when the the master & slave have
 *	not yet been opened (or have since both been closed).
 * Return value = 0 if there is an error 
 * Return value = 1 if no error:
 *	psuMasterName has name of the master device,
 *	psuMaster has its FD
 *	psuSlaveName has name of the corresponding slave device.
 *		(but slave device is not opened yet)
 * (Ownership and permissions of the master port are not changed).
 *----------------------------------------------------------------
 */

#if defined(CLONUS)
static int
openMaster(psu)
servpsu_t *psu;
{
char *pc;

	do_log(IL_DEBUG,"openMaster: clone device %s", CLONE_NAME);

	psu->psuFlag &= ~(PSU_MNAME | PSU_MOPEN | PSU_SNAME | PSU_SLINK);

	if((psu->psuMaster = open(CLONE_NAME, O_RDWR)) < 0)
	{
		do_log(IL_ERROR,"Cannot open clone device %s, error %d", 
			CLONE_NAME, errno);
		psu->psuFlag |= PSU_ERROR;
		return 0;
	}

	strcpy(psu->psuMasterName, CLONE_NAME);

	psu->psuFlag |= PSU_MOPEN;

	if(!(pc = PTSNAME(psu->psuMaster)))
	{
		do_log(IL_ERROR, "Cannot get slave tty name");
		psu->psuFlag |= PSU_ERROR;
		return 0;
	}

	strcpy (psu->psuSlaveName, pc);

	do_log(IL_DEBUG,"openMaster: clone slave %s", pc);

	psu->psuFlag |= PSU_SNAME;

	return 1;
}

#else

#if defined GETPTY	/* IRIX 5.3 and 6.2 support */

#include "fcntl.h"
#include "unistd.h"

static int
openMaster(psu)
servpsu_t *psu;
{
	char *line;
	int filedes[2];

	do_log(IL_DEBUG,"openMaster: using IRIX getpty()");

	psu->psuFlag &= ~(PSU_MNAME | PSU_MOPEN | PSU_SNAME | PSU_SLINK);

	line = _getpty(&filedes[0], O_RDWR, 0600, 0);

	if (line == 0)
	{
		do_log(IL_ERROR,"getpty() failed, error %d", errno);
		psu->psuFlag |= PSU_ERROR;
		return 0;
	}

	psu->psuMaster = filedes[0];

	/* the getpty library routine is using the standard clone device */

	strcpy(psu->psuMasterName, CLONE_NAME);

	do_log(IL_DEBUG,"openMaster: master pty %s", CLONE_NAME);

	psu->psuFlag |= PSU_MOPEN;

	if ((filedes[1] = open(line, O_RDWR)) < 0)
	{
		do_log(IL_ERROR, "Cannot get slave tty name");
		psu->psuFlag |= PSU_ERROR;
		ClosePSU(filedes[0]);
		/* close(filedes[0]); */
		return 0;
	}

	do_log(IL_DEBUG,"openMaster: slave pty %s", line);

	strcpy(psu->psuSlaveName, line);

	return 1;
}


#else	/* Not CLONUS Not GETPTY*/

static char c8tab[] = "pqrsabcdefghijklmnotuvwxyz";
static char c9tab[] = "0123456789abcdef";
static char c10tab[] = "0123456789abcdef";

static int
openMaster(psu)
servpsu_t *psu;
{
char * pc8;
char * pc9;
char * pc10;
char tname[16];
struct stat stat_buff;

	do_log(IL_DEBUG,"openMaster: trying /dev/pty???");

	psu->psuFlag &= ~(PSU_MNAME | PSU_MOPEN | PSU_SNAME | PSU_SLINK);

	strcpy(tname, "/dev/ptyNN");

	for (pc8 = c8tab; *pc8; pc8++)
	{
		tname[8] = *pc8;
		tname[9] = '0';
		tname[10] = 0;
		tname[11] = 0;

		/* If there is no ptyN0, there is no ptyN?, so next N */

		if (stat(tname, &stat_buff) < 0) continue;

		for(pc9 = c9tab; *pc9; pc9++)
		{
			tname[9] = *pc9;
			tname[10] = 0;
			tname[11] = 0;

			if((psu->psuMaster = open(tname, O_RDWR)) >= 0)
			{
				strcpy(psu->psuMasterName, tname);
				do_log(IL_DEBUG,"openMaster: master %s", tname);
				tname[5] = 't';
				strcpy(psu->psuSlaveName, tname);
				psu->psuFlag |= (PSU_SNAME|PSU_MOPEN|PSU_MNAME);
				do_log(IL_DEBUG,"openMaster: slave %s", tname);
				return 1;
			}

			/* lets try these guys too */

			for(pc10 = c10tab; *pc10; pc10++)
			{
				tname[10] = *pc10;
				tname[11] = 0;

				if((psu->psuMaster = open(tname, O_RDWR)) >= 0)
				{
					strcpy(psu->psuMasterName, tname);
					do_log(IL_DEBUG,"openMaster: master %s", tname);
					tname[5] = 't';
					strcpy(psu->psuSlaveName, tname);
					psu->psuFlag |= (PSU_SNAME|PSU_MOPEN|PSU_MNAME);
					do_log(IL_DEBUG,"openMaster: slave %s", tname);
					return 1;
				}
			}
		}
	}	

	do_log(IL_ERROR,"Cannot find a master pseudo-tty");

	psu->psuFlag |= PSU_ERROR;

	return 0;
}

#endif /* not GETPTY */

#endif /* not CLONUS */



/*---------------------------------------------------------------
 * linkSlave(psu)
 * 
 * Prepares the slave for use and links our
 * device to it.
 * Returns 0 if any error
 * Returns 1 if slave is linked
 *---------------------------------------------------------------
 */

static int
linkSlave(psu)
servpsu_t *psu;
{
struct stat st;

/* OS's other than AIX that use streams clone open, support the
 * grantpt() and unlockpt() functions.
 */

#if defined(CLONUS) && defined(IS_GRANT)

	do_log(IL_DEBUG,"linkSlave: Clone");

	/* grantpt forks and execs /usr/lib/pt_chmod, watch out for death of child */
	sighold(SIGCLD);
	if (grantpt(psu->psuMaster) < 0)
	{
		do_log(IL_ERROR, "grantpt failed on %s (%d)", psu->psuMasterName, errno);
		psu->psuFlag |= PSU_ERROR;
		return 0;
	}
	sigrelse(SIGCLD);

	if (unlockpt(psu->psuMaster) < 0)
	{
		do_log(IL_ERROR, "unlockpt failed on %s (%d)", psu->psuMasterName, errno);
		psu->psuFlag |= PSU_ERROR;
		return 0;
	}

/* For AIX and non-streams-clone, we must chown manually, for
 * non-AIX streams clone things, grantpt performs a chown
 * and an initial chmod, but grantpt's chmod will have been too
 * restrictive...
 */

#else	/* Not CLONUS and IS_GRANT */

	do_log(IL_DEBUG,"linkSlave: not Clone");

	chown(psu->psuSlaveName, getuid(), getgid());

#endif	/* Clonus, Not Clonus */


	/* chmod is for _everyone, to guarantee generosity */
	
	chmod(psu->psuSlaveName, 0666);	

#ifdef MAKE_NODE

	/*
	 * rather than do a soft or hard link from our device name
	 * to the pseudo device name, make a node with same major/minor.
	 * hopefully this will get around the ttyname() problem...
	 */

	if (stat(psu->psuSlaveName,&st) < 0)
	{
		do_log(IL_ERROR,"stat(%s) failed, error %d", 
			psu->psuSlaveName, errno);
		return 0;
	}

	/* returns EEXIST if the node already exists */

	unlink(psu->psuLink);		/* some exposure here! */

	if (mknod(psu->psuLink, S_IFCHR | 0666, st.st_rdev) < 0)
	{
		do_log(IL_ERROR,"mknod(%s) failed, error %d", 
			psu->psuLink, errno);
		return 0;
	} 

	/* It appears that the permissions in mknod above are masked
	 * by your umask, so that on irix i only get 644 on the device node
	 * therefore I will do an explicit 666 here.
	 */

	chmod(psu->psuLink, 0666);

	do_log(IL_DEBUG,"linkSlave: mknod %s", psu->psuLink);

#else	/* not MAKE_NODE */

	/* Then we unlink any stale references to our new name, and
	 * create the new link (which for AIX is symbolic)
	 */

	unlink(psu->psuLink);		/* some exposure here! */

	if (LINK(psu->psuSlaveName, psu->psuLink) < 0)
	{
		do_log(IL_ERROR,"link(%s,%s) failed, error %d", 
			psu->psuSlaveName, psu->psuLink, errno);
		return 0;
	}

	do_log(IL_DEBUG,"linkSlave: link %s", psu->psuLink);

#endif	/* MAKE_NODE */

/* For symbolic links like AIX32 uses, we need to chmod both inodes */

#if defined(AIX32)

	chown(psu->psuLink, getuid(), getgid());
	chmod(psu->psuLink, 0666);

#endif	/* AIX32 */

	psu->psuFlag |= PSU_SLINK;
	return 1;
}



/*---------------------------------------------------------------
 * pushSlave(psu)
 *
 * Arranges for the correct modules to
 * be pushed onto the slave port.  
 * There are three possibilities:
 * if autopush is supported (-DPUSH_A)
 * 	then this is done through administration device.
 * if manual push is supported (-DPUSH_M)
 *	then this is done by holding the slave port open,
 *	pushing the modules, and closing the port
 *	immediately after the first data is read from the
 *	master device. When we read data, we know that _he has
 *	it open: so we close it so that his close will be a
 *	last close, so we will get an read of 0 to indicate
 *	that the slave has been closed.
 * if neither PUSH_A nor PUSH_M is used, then PUSH_X will be
 * defined, and this ain't a streams device and there is nothing 
 * to push in any case
 *
 * If you have not defined exactly one of PUSH_?, then there will
 * be a link error because the pushSlave function will be defined
 * multiply or not at all.
 *
 * Return: 1 if successful, 0 if failure
 * 
 * Always returns with the slave port open, even if opening
 * it was not necessary for the pushing.
 *---------------------------------------------------------------
 */

#if defined(PUSH_A)	/* That is, autopush is supported */

static int
pushSlave(psu)
servpsu_t *psu;
{
struct strapush mods;	/* Which modules to push? */
struct stat statbuf;	/* for file information */

	/* If we have done this, we don't need to re-do it */
	if(psu->psuFlag & PSU_DIDAUTO) return 1;

	do_log(IL_DEBUG,"pushSlave: auto");

	if (stat(psu->psuSlaveName, &statbuf) == -1)
	{
		do_log(IL_ERROR, "Cannot stat %s", psu->psuSlaveName);
		psu->psuFlag |= PSU_ERROR;
		return 0;
	}
		
	if ((psu->psuAdm = open_dev_sad_admin(psu)) == -1)
	{
		do_log(IL_ERROR, "Cant open /dev/sad/admin, error %d", errno);
		psu->psuFlag |= PSU_ERROR;
		return 0;	
	}

	/* Clear any modules which might be already present */

	mods.sap_cmd	= SAP_CLEAR;
	psu->psuMajor = mods.sap_major	= major(statbuf.st_rdev);
	psu->psuMinor = mods.sap_minor	= minor(statbuf.st_rdev);
	
	/* If things are already clear, we would expect an ENODEV.
	 * Therefore any _other error would be bad
	 */
	if(ioctl(psu->psuAdm, SAD_SAP, &mods) == -1 && errno != ENODEV)
	{
		do_log(IL_ERROR, "SAP_CLEAR command failed: error %d", errno);
		psu->psuFlag |= PSU_ERROR;
		close(psu->psuAdm);
		return 0;
	}

	mods.sap_cmd	= SAP_ONE;
	mods.sap_npush	= 2;
	strcpy(mods.sap_list[0], "ptem");
	strcpy(mods.sap_list[1], "ldterm");

	if(ioctl(psu->psuAdm, SAD_SAP, &mods) == -1)
	{
		do_log(IL_ERROR, "SAP_ONE command failed: error %d, errno");
		psu->psuFlag |= PSU_ERROR;
		close(psu->psuAdm);
		return 0;
	}

	close(psu->psuAdm);	/* Once we have done the ioctls, we don't
				 * need the adm port open anymore.
				 */

	if((psu->psuSlave = open(psu->psuSlaveName, O_RDWR))   == -1)
	{
		do_log(IL_ERROR, "Cannot open slave %s, error %d", 
			psu->psuSlaveName, errno);
		psu->psuFlag |= PSU_ERROR;
		return 0;
	}
	psu->psuFlag |= PSU_DIDAUTO;

	psu->psuFlag |= (psu->psuFlag & (PSU_PERM | PSU_IGCLOSE)) ?
		PSU_SOPEN : (PSU_SOPEN | PSU_CLEAN);

	return 1;
}
#endif /* PUSH_A */

#if defined(PUSH_M)	/* That is, manual push is used */

static int
pushSlave(psu)
servpsu_t *psu;
{

	do_log(IL_DEBUG,"pushSlave: manual");

	/* Not sure whether it could be already open, but if we
	 * wandered here by accident, at least start clean.
	 */
	if(psu->psuFlag & PSU_SOPEN)
	{
		psu->psuFlag &= ~(PSU_SOPEN | PSU_CLEAN);
		ClosePSU(psu->psuSlave);
		/* close(psu->psuSlave); */
	}

	/* Now open and push all the modules. */

	do_log(IL_DEBUG, "Opening slave %s...", psu->psuSlaveName);
	if((psu->psuSlave = open(psu->psuSlaveName, O_RDWR)) == -1)
	{
		do_log(IL_ERROR, "Cannot open slave %s, error %d", 
			psu->psuSlaveName, errno);
		psu->psuFlag |= PSU_ERROR;
		return 0;
	}

	/* Note, we use \042 to generate double quote because not all
	 * compilers allow doubled double-quotes to represent a single
	 * double-quote within a string, but they all support octal
	 * representation.
	 */
	do_log(IL_DEBUG, "Pushing module ptem...");
	if(ioctl(psu->psuSlave, I_PUSH, "ptem") == -1)
	{
		ClosePSU(psu->psuSlave);
		/* close(psu->psuSlave); */
		do_log(IL_ERROR, "Cannot push module \042ptem\042, error %d",
			errno);
		psu->psuFlag |= PSU_ERROR;
		return 0;
	}
	do_log(IL_DEBUG, "Pushing module ldterm...");
	if(ioctl(psu->psuSlave, I_PUSH, "ldterm") == -1)
	{
		ClosePSU(psu->psuSlave);
		/* close(psu->psuSlave); */
		do_log(IL_ERROR, "Cannot push module \042ldterm\042, error %d",
			errno);
		psu->psuFlag |= PSU_ERROR;
		return 0;
	}

	/* Record the fact that we have the slave port open, and
	 * if we are having a temporary connection, note that we
	 * have to close the port as soon as we see some master input.
	 */

	psu->psuFlag |= (psu->psuFlag & (PSU_IGCLOSE | PSU_PERM)) ?
		PSU_SOPEN : (PSU_SOPEN | PSU_CLEAN);

	return 1;
}

#endif	/* PUSH_M */

#if defined(PUSH_X)	/* That is, no pushing required */

static int
pushSlave(psu)
servpsu_t *psu;
{
	do_log(IL_DEBUG,"pushSlave");

	psu->psuFlag &= ~(PSU_SOPEN | PSU_CLEAN);
	if((psu->psuSlave = open(psu->psuSlaveName, O_RDWR))   == -1)
	{
		do_log(IL_ERROR, "Cannot open slave %s, error %d", 
			psu->psuSlaveName, errno);
		psu->psuFlag |= PSU_ERROR;
		return 0;
	}
	psu->psuFlag |= (psu->psuFlag & (PSU_PERM | PSU_IGCLOSE)) ?
		PSU_SOPEN : (PSU_SOPEN | PSU_CLEAN);

	return 1;
}
#endif	/* PUSH_X */



/*---------------------------------------------------------------
 * uncookPseudo(psu)
 *
 * If raw mode or output-cooked only mode
 * were specified, clears iflag, and lflag on
 * the slave port. If raw specified, clears
 * oflag also, otherwise sets it to ONLCR 
 * and TAB3
 * 
 * Returns 1 if ok, -1 if any error
 *---------------------------------------------------------------
 */

static int
uncookPseudo(psu)
servpsu_t *psu;
{
struct termio my_termio;

	/* If nothing to do, we did it! */
	if (!(psu->psuFlag & (PSU_SRAW | PSU_OCOOK)) ) return 1;

	if (ioctl(psu->psuSlave, TCGETA, &my_termio) < 0)
	{
		do_log(IL_ERROR,"TCGETA failed, error %d", errno);
		return 0;
	}

	my_termio.c_iflag = 0;
	my_termio.c_lflag = 0;
	my_termio.c_oflag = (psu->psuFlag & PSU_OCOOK) ?
		(OPOST | ONLCR | TAB3) : 0 ;

	my_termio.c_cc[VMIN] = 1;
	my_termio.c_cc[VTIME] = 0;

	if (rmt_cmd_flag & REMOTE_MSR_CMD) {
		my_termio.c_cflag &= ~CLOCAL;
	}

	if (ioctl(psu->psuSlave, TCSETA, &my_termio) < 0)
	{
		do_log(IL_ERROR,"TCSETA failed, error %d", errno);
		return 0;
	}

	return 1;
}



/*---------------------------------------------------------------
 * cleanPseudo(psu)
 * 
 * For flavours which do not use autopush,
 * this does nothing.  Otherwise, it eliminates
 * any existing auto-push from that module.
 *---------------------------------------------------------------
 */

int
cleanPseudo(psu)
servpsu_t *psu;
{

#if defined(PUSH_A)	/* That is, if autopush */

struct strapush mods;	/* Which modules to push? */

	if(!(psu->psuFlag & PSU_DIDAUTO)) return 1;

	/* But if DIDAUTO is set, we have set up autopushes, and
	 * we need to clean them off.  We closed the admin port,
	 * so we need to open it up again.
	 */

	if ((psu->psuAdm = open_dev_sad_admin(psu)) == -1)
	{
		do_log(IL_ERROR, "Cannot open /dev/sad/admin, error %d", errno);
		psu->psuFlag |= PSU_ERROR;
		return 0;	
	}

	/* Clear any modules which might be already present */
	mods.sap_cmd	= SAP_CLEAR;
	mods.sap_major	= psu->psuMajor;
	mods.sap_minor	= psu->psuMinor;
	
	if(ioctl(psu->psuAdm, SAD_SAP, &mods) == -1)
	{
		do_log(IL_ERROR, "SAP_CLEAR command failed, error %d",
			errno);
		psu->psuFlag |= PSU_ERROR;
	}
	close(psu->psuAdm);

	/*
	 * Since we've cleared the auto pushes, make sure they're
	 * setup when we re-create the pty.
	 */
	psu->psuFlag &= ~PSU_DIDAUTO;

	return ((psu->psuFlag & PSU_ERROR) ? 0 : 1);

#else	/* not PUSH_A */

	return 1;

#endif	/* PUSH_A */
}



/*---------------------------------------------------------------
 * buildPseudo(psu)
 *
 * If there is an existing pseudoPort 
 * pair, it is torn down.  Then a new
 * set is completely rebuilt.
 *
 * This can therefore be used the first time
 * we want to build a set, and on subsequent times.
 *
 * Caller should ignore SIGHUP signals before calling
 * this.
 *
 * Returns 1 if ok, otherwise 0 if error.
 *---------------------------------------------------------------
 */

int
buildPseudo(psu)
servpsu_t *psu;
{
	do_log(IL_DEBUG,"buildPseudo");

	if (psu->psuFlag & (PSU_SOPEN))
	{
		psu->psuFlag &= ~(PSU_SOPEN | PSU_CLEAN);
		ClosePSU(psu->psuSlave);
		/* close(psu->psuSlave); */
	}

	cleanPseudo(psu); /* knock out any autopush before closing master */

	if (psu->psuFlag & PSU_MOPEN)
	{
		psu->psuFlag &= ~PSU_MOPEN;
		ClosePSU(psu->psuMaster);
		/* close(psu->psuMaster); */
		sleep(2);
	}

	if (!openMaster(psu)) return 0;

	if (!linkSlave(psu)) return 0;

	if (!pushSlave(psu)) return 0;

	if (!uncookPseudo(psu)) return 0;

	return 1;
}		



/*---------------------------------------------------------------
 * reopenSlave(servpsu_t *psu)
 * Will re-open the slave port if it has been
 * closed.
 *---------------------------------------------------------------
 */

int
reopenSlave(psu)
servpsu_t *psu;
{
	if(psu->psuFlag & PSU_SOPEN) return 1;
		
	if (!pushSlave(psu)) return 0;

	if (!uncookPseudo(psu)) return 0;

	return 1;
}


/*---------------------------------------------------------------
 * closeSlave(psu)
 *
 * If slave port is open, will close slave and 
 * return true. Otherwise will return false.
 *---------------------------------------------------------------
 */

int
closeSlave(psu)
servpsu_t *psu;
{
	if(psu->psuFlag & PSU_CLEAN)
	{
		ClosePSU(psu->psuSlave);
		/* close(psu->psuSlave); */
		(psu)->psuFlag &= ~(PSU_SOPEN | PSU_CLEAN);
		return 1;
	}
	else
	{
		return 0;
	}
}


/*---------------------------------------------------------------
 * open_dev_sad_admin(psu)
 *
 * Many svr4 systems only have 8 of these devices so if you are
 * starting many iservds there may not be enough all at once!!!
 *---------------------------------------------------------------
 */

open_dev_sad_admin(psu)
servpsu_t *psu;
{
	int fd;
	int i;

	for (i = 0; i < 30; i++)
	{
		if ((fd = open("/dev/sad/admin", O_RDWR)) == -1 && errno == ENXIO)
		{
			do_log(IL_ERROR, "/dev/sad/admin conflict, retry %d", i);
			sleep(1);
		}

		else return fd;
	}

	return -1;
}



/*---------------------------------------------------------------
 * delPseudo(psu)
 * obviously we can only get here if we are not killed. 
 * at exit, removes the device that was originally created.
 * Returns the result of the unlink.
 *---------------------------------------------------------------
 */

int
delPseudo(psu)
servpsu_t *psu;
{
	int result = 0;

	if (psu->psuFlag & PSU_DELETE)
	{
		result = unlink(psu->psuLink);

		do_log(IL_DEBUG,"delPseudo: unlink %s = %d,%d", psu->psuLink,result,errno);
	}

	return result;
}


/*
 * This will close a pseudo-tty, flushing the i/o queues first.
 */
void
ClosePSU(fd)
int	fd;	/* file descriptor of open pseudo-tty */
{
	int	rc;

	if (rc = tcflow(fd, TCIOFF))	/* Turn input off */
		do_log(IL_DEBUG,"tcflow(%d, TCIOFF)=%d: %d", fd, rc, errno);
	if (rc = tcflow(fd, TCOOFF))	/* Turn output off */
		do_log(IL_DEBUG,"tcflow(%d, TCOOFF)=%d: %d", fd, rc, errno);
	if (rc = tcflush(fd, TCIOFLUSH))	/* Flush all pending I/O */
		do_log(IL_DEBUG,"tcflush(%d, TCIOFLUSH)=%d: %d", fd, rc, errno);
	rc = close(fd);
	do_log(IL_DEBUG,"pseudo-tty close(%d)=%d: %d", fd, rc, errno);
}
