/*---------------------------------------------------------------
 * iservtel.c
 *
 * iservd functions for handling telnet protocol
 *---------------------------------------------------------------
 */
/*
 * $RCSfile: iservtel.c,v $
 * $Revision: 1.1 $
 * $Date: 1998/07/31 21:13:43 $
 *
 * $Header: //Iservd/iservtel.c#2 $
 * 
 * $Log: iservtel.c,v $
 * Revision 1.1  1998/07/31 21:13:43  chuckh
 * Initial revision
 *
 */


/*---------------------------------------------------------------
 * Revision History:
 *
 *  8/05/94	MAG	First Draft
 *---------------------------------------------------------------
 */

#include "isysinc.h"
#include "iservopt.h"
#include "iservlog.h"
#include "iservtel.h"

/*
 * Definitions for the TELNET protocol.
 */

#define	IAC		255
#define	DONT		254	/* Tell target not to  use option */
#define	DO		253	/* Tell target to use option */
#define	WONT		252	/* Tell target that source won't use option */
#define	WILL		251	/* Tell target that source will use option */
#define DATA		1
#define CR		'\r'	/* Carrige return found */
#define TELOPT_BINARY	0	/* Put server in to binary mode */
#define TELOPT_ECHO	1	/* Put server in to binary mode */
#define TELOPT_SGA	3	/* Put server in to binary mode */
#define TELOPT_TM	6	/* wait acknowledge data drained */

#define TEL_BUFF	3	/* 3 byte data message */


/*---------------------------------------------------------------
 * send_telopt(fd, command, telopt_code)
 * 
 * Helper routine to send a TELOPT message 
 * to the server.
 *
 * Returns: -1 if an error
 *---------------------------------------------------------------
 */

static int
send_telopt(fd, command, telopt_code)
int fd;
unsigned char command;
unsigned char telopt_code;
{
	unsigned char message[TEL_BUFF];

	message[0] = IAC;
	message[1] = command;
	message[2] = telopt_code;

	traceData("Telnet Cmd Sent:", message, TEL_BUFF);

	return (write(fd, message, TEL_BUFF));
}



/*---------------------------------------------------------------
 * setBinary(fd)
 *
 * Send TELOPT to set binary mode:
 *
 * Returns -1 if error, >0 if good.
 *---------------------------------------------------------------
 */

int
setBinary(fd)
int fd;
{
	return(send_telopt(fd, DO, TELOPT_BINARY));
}


/*---------------------------------------------------------------
 * wakeTelnet()
 * 
 * Awaken a process waiting in waitTelnet().
 * Any signal handlers responding to connection state
 * should call this.
 *---------------------------------------------------------------
 */

static int stopWaiting;

void
wakeTelnet1()
{
int status;
	stopWaiting = 1;
	wait(&status);
	do_log(IL_DEBUG,"Child died, status %xH", status);
}

void
wakeTelnet2()
{
	stopWaiting = 1;
	do_log(IL_DEBUG,"Telnet drain received");
}

void
wakeTelnet()
{
	stopWaiting = 1;
}



/*---------------------------------------------------------------
 * waitTelnet(fd)
 *
 * Send TELOPT to wait for a response that
 * the data has drained, and wait accordingly.
 *
 * Returns 1 if ok, 
 * Returns 0 if attempt to send the option failed
 *---------------------------------------------------------------
 */

int
waitTelnet(fd)
int fd;
{
	void (*chld)(),(*usr1)();

	stopWaiting = 0;

	chld = signal(SIGCHLD, wakeTelnet1);
	usr1 = signal(SIGUSR1, wakeTelnet2);

	if(send_telopt(fd, DO, TELOPT_TM) < 0) return 0;

	do_log(IL_DEBUG,"Telnet drain sent");

	while(!stopWaiting) 
	{
		pause();
	}

	signal(SIGCHLD, chld);
	signal(SIGUSR1, usr1);

	return 1;
}



/*---------------------------------------------------------------
 * telnetEncode(source, dest, size, binary)
 *
 * Change IAC in data to IAC,IAC sequence
 *---------------------------------------------------------------
 */

int
telnetEncode(source, dest, len, binary)
unsigned char *source;
unsigned char *dest;
int len;
int binary;
{
	unsigned char	*ndx;
	unsigned char	*data;
	telbuf_t	*in_buf;
	unsigned char	c;
	int		tmp;

	tmp = len;	/* for debugging only */

	ndx	= dest;
	in_buf	= (telbuf_t *)source;
	data	= in_buf->tb_buf;

	if (!(rmt_cmd_flag & REMOTE_RTS_CMD)) {
		/* RTS UP/DOWN flag is not set */
		for (; len > 0; len--) 
		{
			if ((*data) == IAC)
			{
				*dest++ = IAC;
			}
			*dest++ = *data;
			/* For compatibility with SUNS */
			if (!binary && (*data == '\r'))
			{
				if (len && *(data+1) != '\n')
					*dest++ = 0; /* Add a NULL */
			}
			data++;
		}
	}
	else {
		/* Flag is set to process RTS up/down commands */

	   while (len > 0) {
		c = *data++;
		len --;
		switch(in_buf->tb_state) {

		case	TS_DATA :
			if (c == IAC) {
				in_buf->tb_state = TS_IAC;
				break;
			}
			*dest++ = c;

			/* For compatibility with SUNS */
			if (!binary && (c == '\r')) {
				if (len && *(data+1) != '\n')
					*dest++ = 0; /* Add a NULL */
			}
			break;
		
		case	TS_IAC :
			*dest++ = IAC;

			switch (c) {
			/*
			 * Command to raise/lower RTS
			 */
			case	RTSUP :
			case	RTSDN :
				*dest++ = c;
				break;

			default :
			/*
			 * IAC in the data path
			 */
				*dest++ = IAC;

				/* One more FF in the data */
				if (c == IAC)
					continue;
				*dest++ = c;				
				break;
			}
			in_buf->tb_state = TS_DATA;
			break;

		default :
			do_log(IL_NOTICE, "Unknown TS_STATE!!!\n");
			break;
		}

	   }
	}
	return(dest - ndx);
}



/*---------------------------------------------------------------
 * telnetDecode(source, dest, len, binary, sockFd)
 *
 * Extract and honor telnet sequences.
 *---------------------------------------------------------------
 */

int
telnetDecode(source, dest, len, binary, sockFd)
unsigned char *source;
unsigned char *dest;
int len;
int binary;
int sockFd;
{
unsigned char *ndx;
int cdata;
static	int state = DATA;
unsigned char response[TEL_BUFF];
int tmp;
		

	tmp = len;	/* for debugging only */
	ndx = dest;

	response[0] = IAC;
	
	while(len > 0)
	{
		cdata = *source++;
		len--; 
		switch(state)
		{
			case DATA:
				switch(cdata)
				{
					case IAC:
						state = IAC;
						break;
					default:
						*dest++ = cdata;
						if (!binary && 
							(cdata == '\r'))
						    state = CR;	
						break;
				}
				break;
			case CR:
					/* remove NULLs after \r's */
					/* Formerly removed \n also, but that seems
					 * unjustified in these circumstances.
					 */
				state = DATA;
				if ( cdata )
					*dest++ = cdata;
				break;
			case IAC:
				switch(cdata)
				{
					case IAC:
						*dest++ = cdata;
						state = DATA;
						break;
					case DO:
					case DONT:
					case WILL:
					case WONT:
						state  = cdata;
						break;
					case MSR:
						do_log(IL_DEBUG, "msr val rcvd");
						if (rmt_cmd_flag & 
							REMOTE_MSR_CMD)
							state = cdata;
						break;
				}
				break;
			case MSR:
				state = DATA;
				/* If the DCD flag is set in the msr value
				 * then signal that to the write process, i.e
				 * the parent
				 */
				if (cdata & MSR_DCD)
					kill(getppid(), SIGUSR2);
				break;
			case DO:
				state = DATA;
				response[1] = DONT;
				response[2] = cdata;
				traceData("Telnet Response Sent:", 
					response, TEL_BUFF);
				write(sockFd, response, TEL_BUFF);
				break;
			case DONT:
				state = DATA;
				break;
			case WILL:
				state = DATA;
			 /* The IntelliServer responded with Will Binary Mode */
				if (binary && (cdata == TELOPT_BINARY))
				{
					break;
				}

				response[1] = DONT;
				if ((cdata == TELOPT_ECHO) ||
					((binary &&(cdata == TELOPT_SGA))))
					response[1] = DO;
				response[2] = cdata;
				traceData("Telnet Response Sent:", 
					response, TEL_BUFF);
				write(sockFd, response, TEL_BUFF);

					/* notify parent received mark*/
				if (cdata == TELOPT_TM)
					kill (getppid (), SIGUSR1); 
				break;
			case WONT:
				state = DATA;
					/* notify parent received mark*/
				if (cdata == TELOPT_TM)
					kill (getppid (), SIGUSR1); 
				break;
		}
	}

#ifdef THIS_CAUSES_UNWANTED_LOGGING
	if (tmp != dest-ndx)
		do_log(IL_DEBUG,"Decoded from %d to %d bytes.", tmp, dest-ndx);
#endif

	return(dest-ndx);
}


